/*
 * Copyright (c) 2002,2004 Nokia
 * Author: lauri.leukkunen@nokia.com
 *
 * sb_gcc_wrapper is licensed under GPL, see COPYING for details.
 *
 * */

#define _GNU_SOURCE

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <libgen.h>



#ifndef SCRATCHBOX_CONFIG
#define SCRATCHBOX_CONFIG "/targets/links/scratchbox.config"
#endif

#ifndef COMPILER_CONFIG
#define COMPILER_CONFIG "/scratchbox/etc/compilers.config"
#endif

#ifdef DEBUG
#define DEBUG_OUT(...) fprintf(stderr, __VA_ARGS__)
#else
#define DEBUG_OUT(...)
#endif

#define BUFFER_SIZE 1024

enum file_groups { COMPILER_GENERIC = 1, COMPILER_GCC, LINKER };

char *compiler_config = COMPILER_CONFIG;
char *scratchbox_config = SCRATCHBOX_CONFIG;
char *default_gcc_prefix = "i386-linux-";


char *gcc_progs[] = {
	"g++",
	"c++",
	"gcc",
	"cpp",
	"cc",
	NULL
};


char *gcc_tools[] = {
	"ld",
	"addr2line",
	"ar",
	"as",
	"gccbug",
	"gcov",
	"nm",
	"objcopy",
	"objdump",
	"ranlib",
	"readelf",
	"size",
	"strings",
	"strip",
	NULL
};


typedef struct compiler_s {
	struct compiler_s *prev;
	struct compiler_s *next;

	char *prefix;
	char *substitute_prefix;
	char *gcc_dir;
	char *specs_file;
	char *ld_args;
} compiler_t;

compiler_t *compiler_list_head = NULL;
compiler_t *compiler_list_tail = NULL;

extern char **environ;

size_t strnlen(const char *str, size_t max_len) 
{
	size_t s=0;
	while (str[s] != '\0' && s < max_len) 
		++s;

	if (s < max_len) return s;
	else return max_len;
}

void death(char *msg, int ret)
{
	fprintf(stderr,"sb_gcc_wrapper: %s\n", msg);
	exit(ret);
}

/*
 * In-place trimming of a string. size is the length of string without trailing '\0'
 */

void strntrim(char *str, int size) 
{
	int ws_count = 0, i;
	if (size <= 0) return;
	while ((str[ws_count] == ' ' || str[ws_count] == '\t' || str[ws_count] == '\n') && ws_count < size) 
		++ws_count;
	for (i=0;i<size - ws_count;++i) {
		str[i] = str[i+ws_count];
	}
	i=size-1;
	while (i >= 0 && (str[i] == ' ' || str[i] == '\t' || str[i] == '\0' || str[i] == '\n')) {
		str[i] = '\0';
		--i;
	}
	/* str[i] = '\0'; */
}

void remove_quotes(char *str, int size)
{
	int ws_count = 0,i;
	while ((str[ws_count] == ' ' || str[ws_count] == '\t' || str[ws_count] == '\n') && ws_count < size)
		++ws_count;
	if (str[ws_count] == '\"' || str[ws_count] == '\'') str[ws_count] = ' '; /* convert beginning quote to space so strntrim chops it off */
	i = size;
	while ((str[i] == ' ' || str[i] == '\t' || str[i] == '\0' || str[i] == '\n') && i >= 0) {
		--i;
	}
	if (i >= 0 && (str[i] == '\"' || str[i] == '\'')) str[i] = ' '; /* convert trailing quote to space */
}

void register_compiler(compiler_t *compiler)
{
	compiler_t *new_compiler = NULL;
	if (!compiler) return;
	DEBUG_OUT("in register_compiler\n");
	new_compiler = (compiler_t *)malloc(sizeof(compiler_t));
	
	if (compiler->prefix) new_compiler->prefix = strdup(compiler->prefix); else new_compiler->prefix = NULL;
	if (compiler->substitute_prefix) new_compiler->substitute_prefix = strdup(compiler->substitute_prefix); else new_compiler->substitute_prefix = NULL;
	if (compiler->gcc_dir) new_compiler->gcc_dir = strdup(compiler->gcc_dir); else new_compiler->gcc_dir = NULL;
	if (compiler->specs_file) new_compiler->specs_file = strdup(compiler->specs_file); else new_compiler->specs_file = NULL;
	if (compiler->ld_args) new_compiler->ld_args = strdup(compiler->ld_args); else new_compiler->ld_args = NULL;

	DEBUG_OUT("COMPILER: prefix: %s, substitute_prefix: %s, gcc_dir: %s, specs_file %s\n", new_compiler->prefix, new_compiler->substitute_prefix, new_compiler->gcc_dir, new_compiler->specs_file);
	if (!compiler_list_head) {
		new_compiler->prev = NULL;
		new_compiler->next = NULL;
		compiler_list_head = new_compiler;
		compiler_list_tail = compiler_list_head;
	} else {
		new_compiler->prev = compiler_list_tail;
		new_compiler->next = NULL;
		compiler_list_tail->next = new_compiler;
		compiler_list_tail = new_compiler;
	}
}

void read_config(void)
{
	FILE *config_file = NULL;
	char *tmp = NULL;
	char *line = (char *)malloc(BUFFER_SIZE+1), *key, *value;
	compiler_t host_gcc, cross_gcc;
	int line_size, i, value_size, key_size;
	
	host_gcc.prefix = NULL;
	host_gcc.substitute_prefix = NULL;
	host_gcc.gcc_dir = NULL;
	host_gcc.specs_file = NULL;
	host_gcc.ld_args = NULL;

	cross_gcc.prefix = NULL;
	cross_gcc.substitute_prefix = NULL;
	cross_gcc.gcc_dir = NULL;
	cross_gcc.specs_file = NULL;
	cross_gcc.ld_args = NULL;

	tmp = getenv("SBOX_SCRATCHBOX_CONFIG");
	if (tmp) scratchbox_config = strdup(tmp);

	if ((config_file = fopen(scratchbox_config,"r")) == NULL) {
		/* we're unable to open the file we'll die */
		death("Unable to open scratchbox.config",1);
	}
	DEBUG_OUT("default_gcc_prefix before: %s\n", default_gcc_prefix);
	default_gcc_prefix = NULL;
	while (1) {
		if ((fgets(line, BUFFER_SIZE, config_file)) == NULL) {
			death("Error reading scratchbox.config file",1);
		}

		line_size = strlen(line);
		if (line_size == 0) continue;
		strntrim(line, line_size);
		line_size = strlen(line);
		i = 0;
		while (line[i] != '=' && (line[i] && i < line_size)) {
			++i;
		}
		if (i >= line_size) continue;
		key = (char *)malloc(i+1); /* space for trailing '\0' */
		strncpy(key,line,i); /* we don't want the '=' character */
		key[i] = '\0';
		key_size = i;
		value_size = line_size - i - 1; /* don't count the ending '\0' */
		value = (char *)malloc(value_size + 1);
		strncpy(value, &line[i+1], value_size);
		value[value_size] = '\0';
		remove_quotes(value, value_size);
		strntrim(value, value_size);
		value_size = strlen(value);
		if (strcmp(key, "SBOX_DEFAULT_GCC_PREFIX") == 0) {
			default_gcc_prefix = strdup(value);
		} else if (strcmp(key, "SBOX_CROSS_GCC_PREFIX") == 0) {
			cross_gcc.prefix = strdup(value);
		} else if (strcmp(key, "SBOX_CROSS_GCC_SUBST_PREFIX") == 0) {
			cross_gcc.substitute_prefix = strdup(value);
		} else if (strcmp(key, "SBOX_CROSS_GCC_SPECS_FILE") == 0) {
			cross_gcc.specs_file = strdup(value);
		} else if (strcmp(key, "SBOX_CROSS_GCC_DIR") == 0) {
			cross_gcc.gcc_dir = strdup(value);
		} else if (strcmp(key, "SBOX_CROSS_GCC_LD_ARGS") == 0) {
			cross_gcc.ld_args = strdup(value);
		} else if (strcmp(key, "SBOX_HOST_GCC_PREFIX") == 0) {
			host_gcc.prefix = strdup(value);
		} else if (strcmp(key, "SBOX_HOST_GCC_SUBST_PREFIX") == 0) {
			host_gcc.substitute_prefix = strdup(value);
		} else if (strcmp(key, "SBOX_HOST_GCC_SPECS_FILE") == 0) {
			host_gcc.specs_file = strdup(value);
		} else if (strcmp(key, "SBOX_HOST_GCC_DIR") == 0) {
			host_gcc.gcc_dir = strdup(value);
		} else if (strcmp(key, "SBOX_HOST_GCC_LD_ARGS") == 0) {
			host_gcc.ld_args = strdup(value);
		}
		free(key); key = NULL;
		free(value); value = NULL;
		if (default_gcc_prefix && cross_gcc.prefix && cross_gcc.substitute_prefix && cross_gcc.specs_file && cross_gcc.gcc_dir && cross_gcc.ld_args
			&& host_gcc.prefix && host_gcc.substitute_prefix && host_gcc.specs_file && host_gcc.gcc_dir && host_gcc.ld_args) {
			register_compiler(&cross_gcc);
			register_compiler(&host_gcc);
			break;
		}
	}

	/* allow overriding what executing 'gcc' does */
	tmp = getenv("SBOX_DEFAULT_GCC_PREFIX");
	if (tmp) {
		if (default_gcc_prefix) free(default_gcc_prefix);
		default_gcc_prefix = strdup(tmp);
	}
	DEBUG_OUT("default_gcc_prefix after: %s\n",default_gcc_prefix);
	fclose(config_file);
	free(line); line = NULL;
}



/*
 * detect_executable takes filename and breaks it into prefix and progname
 * and returns the executable group
 * */

int detect_executable(char *filename, char **prefix, char **progname)
{
	char *wrk = NULL, *tmp = NULL;
	compiler_t *gcc = NULL;
	int i = 0, filename_length, wrk_length, prefix_length;
	
	if (!filename) return 0;

	DEBUG_OUT("detect_executable: filename: %s\n", filename);
	
	/* check if one of the gcc program names: gcc, g++, cc, c++, cpp */

	filename_length = strlen(filename);
	i=0;
	wrk = gcc_progs[i++];
	while (wrk) {
		DEBUG_OUT("looping in outer while(wrk)\n");
		wrk_length = strlen(wrk);
		DEBUG_OUT("wrk_length, filename_length: %i,%i\n",wrk_length, filename_length);
		/* could it be without prefix? */
		if (wrk_length == filename_length) {
			/* at least possible */
			if (strcmp(wrk,filename) == 0) {
				/* yesh, found a match. now choose the default gcc */
				*prefix = (char *)strdup(default_gcc_prefix);
				*progname = (char *)strdup(filename);
				return COMPILER_GCC;
			} 
		}
		gcc = compiler_list_head;
		while (gcc) {
			DEBUG_OUT("looping in inner while (gcc): gcc->prefix=%s\n", gcc->prefix);
			prefix_length = strlen(gcc->prefix);
			if (wrk_length + prefix_length != filename_length) {
				gcc = gcc->next;
				continue; /* obvious case */
			}
			tmp = (char *)malloc(wrk_length + prefix_length + 1);
			strcpy(tmp,gcc->prefix);
			strcat(tmp,wrk);
			if (strcmp(tmp, filename) == 0) {
				/* match */
				*prefix = (char *)strdup(gcc->prefix);
				*progname = (char *)strdup(wrk);
				free(tmp);
				return COMPILER_GCC;
			}
			free(tmp);
			gcc = gcc->next;
		}
		wrk = gcc_progs[i++];
	}
	
	/* so it wasn't a gcc prog, maybe it is still one of the tools (binutils: ld, addr2line, ar, as, gccbug, gcov ...) */
	i = 0;
	wrk = gcc_tools[i];
	while (wrk) {
		wrk_length = strlen(wrk);
		/* could it be without prefix? */
		if (wrk_length == filename_length) {
			/* at least possible */
			if (strcmp(wrk,filename) == 0) {
				/* yesh, found a match. now choose the default gcc */
				*prefix = (char *)strdup(default_gcc_prefix);
				*progname = (char *)strdup(filename);
				if (strcmp(wrk, "ld") == 0) return LINKER;
				else return COMPILER_GENERIC;
			} 
		}
		gcc = compiler_list_head;
		while (gcc) {
			prefix_length = strlen(gcc->prefix);
			if (wrk_length + prefix_length != filename_length) {
				gcc = gcc->next;
				continue; /* obvious case */
			}
			tmp = (char *)malloc(wrk_length + prefix_length + 1);
			strcpy(tmp,gcc->prefix);
			strcat(tmp,wrk);
			if (strcmp(tmp, filename) == 0) {
				/* match */
				*prefix = (char *)strdup(gcc->prefix);
				*progname = (char *)strdup(wrk);
				free(tmp);
				if (strcmp(wrk, "ld") == 0) return LINKER;
				else return COMPILER_GENERIC;
			}
			free(tmp);
			gcc = gcc->next;
		}
		wrk = gcc_tools[i++];
	}

	/* oh well, give up */

	return -1;
}


compiler_t * find_compiler(char *prefix)
{
	compiler_t *wrk = compiler_list_head;
	while (wrk) {
		if (strcmp(wrk->prefix, prefix) == 0) return wrk;
		wrk = wrk->next;
	}
	return NULL;
}


int build_argv(char *args, char ***argv) 
{
	int c = 0, in_arg = 0, i;
	char *ch, *begin, *end;

	if (!args) return -1;
	
	/* first count them */
	
	ch = args;
	
	while (*ch) {
		if (*ch != ' ' && *ch != '\t' && *ch != '\n') {
			if (!in_arg) {
				in_arg = 1;
				++c;
			}
		} else {
			if (in_arg) {
				in_arg = 0;
			}
		}
		++ch;
	}
	
	/* then copy them to argv */
	
	*argv = (char **)malloc(sizeof(char *) * c + 1);
	in_arg = 0;
	begin = NULL;
	end = NULL;
	i = 0;
	ch = args;
	while (*ch) {
		if (*ch != ' ' && *ch != '\t' && *ch != '\n') {
			if (!in_arg) {
				in_arg = 1;
				begin = ch;
			}
		} else {
			if (in_arg) {
				in_arg = 0;
				end = ch;
				(*argv)[i] = strndup(begin, end - begin);
				++i;
			}
		}
		++ch;
	}
	if (in_arg) (*argv)[i++] = strdup(begin); /* while() loop won't catch the last arg */
	(*argv)[i] = NULL; /* end of list */

	return c;
}

/*
 * exec_program calls execvp and thus should never return
 * */

void exec_program(int argc, char **argv)
{
	char *tmp = NULL;
	char *prefix = NULL;
	char *progname = NULL;
	compiler_t *compiler = NULL;
	char *exec_filename = NULL;
	char *exec_basename = NULL;
	char **exec_argv = NULL;
	char *specs_arg = NULL;
	char *path = NULL;
	int argv_index=1, i = 0, have_specs=0, exec_group=0;
	int exec_filename_size = 0;
	int exec_basename_size = 0;
	int specs_arg_size = 0;
	
	exec_basename = strdup(basename(argv[0]));
	exec_basename_size = strlen(exec_basename);
	
	exec_group = detect_executable(exec_basename, &prefix, &progname);
	if (exec_group < 0) death("Unknown executable, maybe you are trying to run a compiler of a wrong architecture?",1);
	DEBUG_OUT("exec_group: %i, prefix: %s, progname: %s\n",exec_group, prefix, progname);
	compiler = find_compiler(prefix);
	if (!compiler) death("Unknown compiler prefix",1);
	DEBUG_OUT("found a compiler: %s\n", compiler->gcc_dir);

	exec_filename_size = strlen(compiler->gcc_dir) + 1 + strlen(compiler->substitute_prefix) + strlen(progname) + 1;
	exec_filename = (char *)malloc(exec_filename_size);

	tmp = getenv("PATH");
	if (!tmp) {
		path = strdup(compiler->gcc_dir);
	} else {
		path = malloc(strlen(tmp) + 1 + strlen(compiler->gcc_dir) + 1);
		strcpy(path,compiler->gcc_dir);
		strcat(path,":");
		strcat(path,tmp);
	}
	setenv("PATH",path,1);
	DEBUG_OUT("setting PATH: %s\n", path);

	strcpy(exec_filename, compiler->gcc_dir);
	strcat(exec_filename, "/");
	strcat(exec_filename, compiler->substitute_prefix);
	strcat(exec_filename, progname);

	if (strlen(compiler->specs_file) > 0) have_specs=1;

	if ( have_specs && (exec_group == COMPILER_GCC) ) {
		/* handler for all gcc work-alikes */
		int bonus=1, wrk=0;

		DEBUG_OUT("This is a GCC work-alike\n");
		if (have_specs) ++bonus;

		specs_arg_size = strlen("-specs=") + strlen(compiler->specs_file) + 1;
		specs_arg = (char *)malloc(specs_arg_size);
		strcpy(specs_arg,"-specs=");
		strcat(specs_arg, compiler->specs_file);
		exec_argv = (char **)malloc(sizeof(char *) * (argc + bonus));
		exec_argv[wrk++] = exec_filename;
		if (have_specs) exec_argv[wrk++] = specs_arg;
		if (bonus == 2) DEBUG_OUT("wrapper: %s %s\n", exec_argv[0], exec_argv[1]);
		argv_index = bonus;
	} else if (exec_group == COMPILER_GENERIC || exec_group == COMPILER_GCC || exec_group == LINKER) {
		/* default handler for all other compiler binaries and also for
		 * ld and gcc in case we lack linker script and/or spec file.
		 * */
		int bonus = 1, ld_argc=0, ld_i, wrk, alloc_size;
		char **ld_argv;

		DEBUG_OUT("This is a COMPILER_GENERIC\n");
		
		wrk = 0;
		if (exec_group == LINKER) {
			DEBUG_OUT("compiler->ld_args: %s\n", compiler->ld_args);
			ld_argc = build_argv(compiler->ld_args, &ld_argv);
			bonus += ld_argc;
		}
		DEBUG_OUT("argc + bonus: %i + %i\n", argc, bonus);
		alloc_size = sizeof(char *) * (argc + bonus);
		DEBUG_OUT("alloc_size: %i\n", alloc_size);
		exec_argv = (char **)malloc(alloc_size);
		DEBUG_OUT("malloc ok\n");
		exec_argv[wrk++] = exec_filename;
		
		if (exec_group == LINKER) {
			for (ld_i=0;ld_i<ld_argc;++ld_i) {
				exec_argv[wrk++] = ld_argv[ld_i];
			}
		}
		argv_index = bonus;
		DEBUG_OUT("COMPILER_GENERIC finished\n");
	} else {
		/* unknown, bail out */
		death("Unknown executable filename",1);
	}
	
	/* fill in the remainder of the arguments */

	for (i = 1; i < argc; ++i) {
		DEBUG_OUT("adding argument [%i,%i]: %s", argv_index, i, argv[i]);
		exec_argv[argv_index] = argv[i];
		++argv_index;
	}
	exec_argv[argv_index] = NULL;
	
	free(prefix); prefix = NULL;
	
	/* Quick hack for ccache */
	if ((getenv("SBOX_USE_CCACHE")) && (exec_group == COMPILER_GCC)) {
		exec_filename = "/bin/ccache";
	}
	
	if (execve(exec_filename,exec_argv,environ) < 0) {
		DEBUG_OUT("sb_gcc_wrapper: exec_filename: %s\n",exec_filename);
		death("error executing program",5);
	}
}

int main(int argc, char **argv)
{
	/* first deal with the config files */
	read_config();
	if (compiler_list_head) {
		DEBUG_OUT("sb_gcc_wrapper: compilers found!\n");
	} else {
		DEBUG_OUT("sb_gcc_wrapper: fishy\n");
		death("Check your scratchbox.conf file.",1);
	}

	/* Now figure out who we are supposed to be.
	 * Currently supported:
	 *
	 * addr2line, ar, as, c++, c++filt, cpp, g++, gcc, gccbug, gcov, ld
	 * nm, objcopy, objdump, ranlib, readelf, size, strings, strip 
	 * 
	 * of these g++, c++, gcc get special treatment
	 * */

	//DEBUG_OUT("I am: %s\n",argv[0]);

	// exec_program will never return
	exec_program(argc, argv);
	
	return(10);
}
